/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      WeightedSot.cpp
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/* --------------------------------------------------------------------- */
/* --- INCLUDE --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/* SOT */
#include <sot-core/weighted-sot.h>
#include <sot-core/memory-task-sot.h>
#include <sot-core/pool.h>
#include <sot-core/task.h>
#include <sot-core/debug.h>
using namespace std;
using namespace sot;


/* --------------------------------------------------------------------- */
/* --- CLASS ----------------------------------------------------------- */
/* --------------------------------------------------------------------- */


#include <sot-core/factory.h>
SOT_FACTORY_ENTITY_PLUGIN(WeightedSot,"WSOT");

/* --------------------------------------------------------------------- */
/* --- CONSTRUCTION ---------------------------------------------------- */
/* --------------------------------------------------------------------- */
WeightedSot::
WeightedSot( const std::string& name )
  :Sot(name)
   ,weightSIN( NULL,"sotWeightedSOT("+name+")::input(matrix)::weight" )
   ,constrainedWeightSOUT( boost::bind(&WeightedSot::computeConstrainedWeight,
				       this,_1,_2),
			   weightSIN<<constraintSOUT,
			   "sotWeightedSOT("+name+")::input(matrix)::KweightOUT" )
   ,constrainedWeightSIN( NULL,"sotWeightedSOT("+name+")::input(matrix)::Kweight" )
   ,squareRootInvWeightSOUT( boost::bind(&WeightedSot::computeSquareRootInvWeight,this,_1,_2),
			     weightSIN<<constrainedWeightSIN,
			     "sotWeightedSOT("+name+")::output(matrix)::sqweight" )
   ,squareRootInvWeightSIN( &squareRootInvWeightSOUT,
			    "sotWeightedSOT("+name+")::input(matrix)::sqweightIN" )
{
  sotDEBUGIN(25);
  signalRegistration( weightSIN<<constrainedWeightSIN<<squareRootInvWeightSOUT
		      << squareRootInvWeightSIN );

  controlSOUT.setFunction( boost::bind(&WeightedSot::computeWeightedControlLaw,
				       this,_1,_2) );
  controlSOUT.addDependancy( squareRootInvWeightSIN );

  constrainedWeightSIN.plug(&constrainedWeightSOUT);
  sotDEBUGOUT(25);
}


/* --------------------------------------------------------------------- */
/* --- A inv ----------------------------------------------------------- */
/* --------------------------------------------------------------------- */

#include <MatrixAbstractLayer/boostspecific.h>

ml::Matrix& WeightedSot::
computeSquareRootInvWeight( ml::Matrix& S5i,const int& time )
{
  sotDEBUGIN(15);

  const ml::Matrix& A = constrainedWeightSIN( time );
  if( A.nbCols()!= A.nbRows() )
    {
      SOT_THROW ExceptionTask( ExceptionTask::MATRIX_SIZE,
				  "Weight matrix should be square.","" );
    }
  sotDEBUG(25) << "KA = " << A << endl;

  /* Decomposition of Choleski:
   *  (1) Diag of S5 is the sqrt of diag of A minus the sum-square
   * of the begining of colum i of S5: s_ii = sqrt(a_ii - sum(k=0:i-1) s_ki^2).
   *  (2) For each element of the line, the component s_ij is the cor. element
   * of A minus the scalar product of column i and j of s, divided by the diag
   * of S: s_ij = (a_ij - sum(k=0:i) s_ik.s_jk ) / s_ii .
   */
  ml::Matrix S5( A.nbRows(),A.nbRows() ); S5.fill(0.);
  //S5.resize( A.nbRows(),A.nbRows() ); S5.fill(0.);
  for( unsigned int i=0;i<A.nbCols();++i )
    {
      double x=A(i,i);
      for( unsigned int k=0;k<i;++k )
	x-=S5(k,i)*S5(k,i);
      double sq=sqrt(x);
      S5(i,i)=sq;

      for( unsigned int j=i+1;j<A.nbCols();++j )
	{
	  x=A(i,j);
	  for( unsigned int k=0;k<i;++k )
	    x-=S5(k,i)*S5(k,j);
	  S5(i,j)=x/sq;
	}
    }
  sotDEBUG(20) << "S = " <<S5  << endl;

  /* S=Inversion of a upper-triangular matrix A.
   *  (1) the diag of the inverse is the inverse of the diag: s_ii = 1/a_ii.
   *
   *  (2) Then cover the matrix by writting the lines from the diag.
   * For each element i,j, knowing all elements i,k (k<j) of the line i,
   * the element s_ij is obtained by developing the term p_ij of the product
   * S*A. p_ik = [S*A]_ik = 0 = sum(k=i:j) s_ik a_kj = 0. Then
   * s_ij = - [ sum(k=i:j-1) s_ik a_kj ] / a_ii
   */
  S5i.resize( A.nbRows(),A.nbRows() ); S5i.fill(0.);
  for( unsigned int l=0;l<A.nbRows();++l )
    for( unsigned int i=0;i<A.nbCols();++i )
      {
 	unsigned int j=i+l; if(j>=A.nbCols()) continue;
 	if( j==i ) S5i(j,j)=1/S5(j,j);
 	else
 	  {
 	    S5i(i,j)=0;
 	    for( unsigned int k=i;k<j;k++ )
 	      {	S5i(i,j) -= S5(k,j)*S5i(i,k);    }
 	    S5i(i,j)/=S5(j,j);
 	  }
      }
  sotDEBUG(15) << "Si = " <<S5i  << endl;

  sotDEBUGOUT(15);
  return S5i;
}


/* --------------------------------------------------------------------- */
/* --- CONTROL --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

ml::Matrix& WeightedSot::
computeConstrainedWeight( ml::Matrix& KAK,const int& iterTime )
{
  sotDEBUGIN(15);

  const ml::Matrix &K = constraintSOUT(iterTime);
  const ml::Matrix &A = weightSIN(iterTime);
  const unsigned int mJ = K.nbCols();

//   const unsigned int rhand=28, lhand=40;
//   for( unsigned int i=0;i<6;++i )
//     for( unsigned int j=0;j<mJ+6;++j )
//       {
//  	if( i+rhand==j ) A(j,j)=1; else
// 	  A(i+rhand,j) = A(j,rhand+i) = 0.;
//  	if( i+lhand==j ) A(j,j)=1; else
// 	  A(i+lhand,j) = A(j,lhand+i) = 0.;
//       }

  sotDEBUG(35) << "Kff = " << K  << endl;
  sotDEBUG(35) << "A = " << A  << endl;

  KAK.resize( mJ,mJ ); KAK.fill(0);
  {
    ml::Matrix KA( mJ,6 ); KA.fill( 0. );
    for( unsigned int i=0;i<mJ;++i )
      for( unsigned int j=0;j<6;++j )
	for( unsigned int k=0;k<6;++k )
	  KA(i,j) += K(k,i)*A(k,j);

    for( unsigned int i=0;i<mJ;++i )
      for( unsigned int j=0;j<mJ;++j )
	{
	  for( unsigned int k=0;k<6;++k )
 	    {
	      KAK(i,j) += KA(i,k)*K(k,j);
	      const double x = A(i+6,k)*K(k,j);
	      KAK(i,j) += x;
	      KAK(j,i) += x;
	    }
	  KAK(i,j) += A(i+6,j+6);
	}
  }
  sotDEBUG(35) << "KAK = " << KAK  << endl;
  return KAK;
}
/* --------------------------------------------------------------------- */
/* --- CONTROL --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

#ifndef WIN32
#include <sys/time.h>
#endif /*WIN32*/

//#define WITH_CHRONO

#ifdef  WITH_CHRONO
#   define sotINIT_CHRONO1 struct timeval t0,t1; double dt
#   define sotSTART_CHRONO1  gettimeofday(&t0,NULL)
#   define sotCHRONO1 \
      gettimeofday(&t1,NULL);\
      dt = ( (t1.tv_sec-t0.tv_sec) * 1000.\
	     + (t1.tv_usec-t0.tv_usec+0.) / 1000. );\
      sotDEBUG(1) << "dt: "<< dt
#   define sotINITPARTCOUNTERS  struct timeval tpart0
#   define sotSTARTPARTCOUNTERS  gettimeofday(&tpart0,NULL)
#   define sotCOUNTER(nbc1,nbc2) \
	  gettimeofday(&tpart##nbc2,NULL); \
	  dt##nbc2 += ( (tpart##nbc2.tv_sec-tpart##nbc1.tv_sec) * 1000. \
		   + (tpart##nbc2.tv_usec-tpart##nbc1.tv_usec+0.) / 1000. )
#   define sotINITCOUNTER(nbc1) \
   struct timeval tpart##nbc1; double dt##nbc1=0;
#else // #ifdef  WITH_CHRONO
#   define sotINIT_CHRONO1
#   define sotSTART_CHRONO1
#   define sotCHRONO1 if(1) ; else std::cout
#   define sotINITPARTCOUNTERS
#   define sotSTARTPARTCOUNTERS
#   define sotCOUNTER(nbc1,nbc2)
#   define sotINITCOUNTER(nbc1)
#endif // #ifdef  WITH_CHRONO



ml::Vector& WeightedSot::
computeWeightedControlLaw( ml::Vector& control,const int& iterTime )
{
  sotDEBUGIN(15);
  sotDEBUGF(5, " --- Time %d -------------------", iterTime );

  sotINIT_CHRONO1; sotINITPARTCOUNTERS;
  sotINITCOUNTER(0b);
  sotINITCOUNTER(1); sotINITCOUNTER(2);sotINITCOUNTER(3);sotINITCOUNTER(4);
  sotINITCOUNTER(5); sotINITCOUNTER(6);sotINITCOUNTER(7);
  sotINITCOUNTER(8); sotINITCOUNTER(9);
  sotINITCOUNTER(4a); sotINITCOUNTER(4b); sotINITCOUNTER(4c); sotINITCOUNTER(4d);sotINITCOUNTER(4e);

  sotSTART_CHRONO1;
  sotSTARTPARTCOUNTERS;

  ml::Matrix P;
  const double &th = inversionThresholdSIN(iterTime);
  const ml::Matrix &K = constraintSOUT(iterTime);
  const unsigned int mJ = K.nbCols();
  //ml::Matrix A = weightSIN(iterTime);
  const ml::Matrix &S5i = squareRootInvWeightSIN(iterTime);
  ml::Matrix Ai(mJ,mJ);
  S5i.multiply(S5i.transpose(),Ai); // TODO: Optimize by considering the triangular shape!
  sotDEBUG(35) << "Ai = " << Ai << endl;

  /* --- Q0 --- */
  /* --- Q0 --- */
  try {
    control = q0SIN( iterTime );
    sotDEBUG(15) << "initial velocity q0 = " << control << endl;
    if( mJ!=control.size() ) { control.resize( mJ ); control.fill(.0); }
  }
  catch (...)
    {
      if( mJ!=control.size() ) { control.resize( mJ ); }
      control.fill(0.);
      sotDEBUG(25) << "No initial velocity." <<endl;
    }


  /* --- Main Loop --- */
  /* --- Main Loop --- */
  unsigned int iterTask = 0;
  for( StackType::iterator iter = stack.begin(); iter!=stack.end();++iter )
    {
      sotCOUNTER(0,0b); // Direct Dynamic
      sotDEBUGF(5,"Rank %d.",iterTask);
      TaskAbstract & task = **iter;
      const ml::Matrix &JacRO = task.jacobianSOUT(iterTime);
      const ml::Vector err = Sot::taskVectorToMlVector(task.taskSOUT(iterTime));
      const unsigned int nJ = JacRO.nbRows();
      sotCOUNTER(0b,1); // Direct Dynamic

      /* Init memory. */
      MemoryTaskSOT * mem = dynamic_cast<MemoryTaskSOT *>( task.memoryInternal );
      if( NULL==mem )
        {
          if( NULL!=task.memoryInternal ) delete task.memoryInternal;
          mem = new MemoryTaskSOT( task.getName()+"_memSOT",nJ,mJ );
          task.memoryInternal = mem;
        }

      ml::Matrix Jp,V; unsigned int rankJ;
      ml::Matrix Jac(nJ,mJ);
      ml::Matrix Jt(nJ,mJ);

      if( (task.jacobianSOUT.getTime()>mem->jacobianInvSINOUT.getTime())
	  ||(mem->jacobianInvSINOUT.accessCopy().nbRows()!=mJ)
	  ||(mem->jacobianInvSINOUT.accessCopy().nbCols()!=nJ)
	  ||(task.jacobianSOUT.getTime()>mem->jacobianConstrainedSINOUT.getTime())
	  ||(task.jacobianSOUT.getTime()>mem->rankSINOUT.getTime())
	  ||(task.jacobianSOUT.getTime()>mem->singularBaseImageSINOUT.getTime()) )
      {
	sotDEBUG(15) <<"Recompute inverse."<<endl;

	/* --- FIRST ALLOCS --- */
	const unsigned int FF_SIZE=JacRO.nbCols()-mJ;
	ml::Matrix Jff( nJ,FF_SIZE );
	ml::Matrix Jact( nJ,mJ );
	Jp.resize( mJ,nJ );
	/***/sotCOUNTER(1,2); // first allocs

	/* --- COMPUTE JK --- */
	for( unsigned int i=0;i<nJ;++i )
	  {
	    for( unsigned int j=0;j<FF_SIZE;++j ) Jff(i,j)=JacRO(i,j);
	    for( unsigned int j=FF_SIZE;j<JacRO.nbCols();++j )
	      Jact(i,j-FF_SIZE)=JacRO(i,j);
	  }
	Jff.multiply(K,Jac);
	Jac+=Jact;
	/***/sotCOUNTER(2,3); // compute JK
	
	/* --- COMPUTE Jt --- */
	if( mJ==P.nbCols() ) Jac.multiply(P,Jt);
	else { Jt = Jac; }
	/***/sotCOUNTER(3,4); // compute Jt
	
	/* --- COMPUTE S --- */	
	ml::Matrix Kact(mJ,mJ); Kact=S5i;
	Task* taskSpec = dynamic_cast<Task*>( &task );
	if( NULL!=taskSpec )
	  {
	    const Flags& controlSelec = taskSpec->controlSelectionSIN( iterTime );
	    sotDEBUG(25) << "Control selection = " << controlSelec <<endl;
	    if( controlSelec )
	      {
		if(!controlSelec)
		  {
		    /***/sotCOUNTER(4,4a); // compute Kh
		    sotDEBUG(15) << "Control selection."<<endl;
		    std::vector<unsigned int> unactiveList;
		    sotDEBUG(25) << "selec"<<iterTask<<" = [ ";
		    for( unsigned int i=0;i<mJ;++i )
		      {
			if(! controlSelec(i) )
			  {
			    if(sotDEBUG_ENABLE(25)) sotDEBUGFLOW.outputbuffer << i <<" ";
			    unactiveList.push_back(i);
			  }
		      }
		    if(sotDEBUG_ENABLE(25)) sotDEBUGFLOW.outputbuffer << "]" << endl ;
		
		    const unsigned int unactiveSize = unactiveList.size();
		    /***/sotCOUNTER(4a,4b); // compute Kh
	
		    /* Q = H*Ai*H, H being the unactivation matrix. */
		    ml::Matrix Q(unactiveSize,unactiveSize);
		    ml::Matrix Sir(unactiveSize,mJ);
		    for( unsigned int i=0;i<unactiveSize;++i )
		     {
		       for( unsigned int j=0;j<unactiveSize;++j )
			 Q(i,j)=Ai(unactiveList[i],unactiveList[j]);
		
		       for( unsigned int j=0;j<mJ;++j )
			 Sir(i,j) = S5i(unactiveList[i],j);
		     }
		    sotDEBUG(25) << "Q"<<iterTask<<" = " << Q << endl;
		    sotDEBUG(25) << "Sr"<<iterTask<<" = " << Sir << endl;
		    /***/sotCOUNTER(4b,4c); // compute Kh

		    /* Qi = inv(Q), always positive. */
		    ml::Matrix Qi(unactiveSize,unactiveSize);
		    Q.inverse(Qi);
		    sotDEBUG(25) << "Qi"<<iterTask<<" = " << Qi << endl;
		    /***/sotCOUNTER(4c,4d); // compute Kh

		    /* Kact = (HSi)^+ HSi = Si_red^T Qi Si_red. */
		    ml::Matrix SrQi(mJ,unactiveSize);
		    ml::Matrix SrQiSr(mJ,mJ);
		    ml::Matrix ArQiSr(mJ,mJ);
		    Sir.transpose().multiply(Qi,SrQi);
		    SrQi.multiply(Sir,SrQiSr);
		    S5i.multiply(SrQiSr,ArQiSr);
		    Kact-=ArQiSr; // TODO: optimizer ce merdier! Kact est sparse!

		    sotDEBUG(15) << "Kact"<<iterTask<<" = "<<Kact<<endl;
		    /***/sotCOUNTER(4d,4e); // compute Kh
		  }
		else
		  {
		    sotDEBUG(15) << "S is equal to Id."<<endl;
		  }
	      }
	    else
	      {
		sotDEBUG(15) << "Task not activated."<<endl;
		Jt *= 0;
	      }
	  }
	/***/sotCOUNTER(4,5); // compute Kh

	/* --- PINV --- */
	ml::Matrix U; ml::Vector S;
	ml::Matrix JtA( Jt.nbRows(),Jt.nbCols() );
	ml::Matrix JtAp( Jt.nbCols(),Jt.nbRows() );

	Jt.multiply( Kact,JtA );
	JtA.pseudoInverse( JtAp,th,&U,&S,&V );
	S5i.multiply( JtAp,Jp ); // TODO: S5i*JtAp = Kact*JtAp, and Kact is sparse.

	if(  mJ==P.nbCols() ) Jp = P*Jp; // P is not idempotent
	/***/sotCOUNTER(5,6); // PINV

	/* --- RANK --- */
	{
	  const unsigned int Jmax = S.size(); rankJ=0;
	  for( unsigned i=0;i<Jmax;++i ) { if( S(i)>th ) rankJ++; }
	}

	sotDEBUG(45) << "control"<<iterTask<<" = "<<control<<endl;
	sotDEBUG(25) << "J"<<iterTask<<" = "<<JacRO<<endl;
	sotDEBUG(25) << "JK"<<iterTask<<" = "<<Jac<<endl;
	sotDEBUG(25) << "Jt"<<iterTask<<" = "<<Jt<<endl;
	sotDEBUG(25) << "JtA"<<iterTask<<" = "<<JtA<<endl;
	sotDEBUG(15) << "Jp"<<iterTask<<" = "<<Jp<<endl;
	sotDEBUG(15) << "e"<<iterTask<<" = "<<err<<endl;
	sotDEBUG(45) << "JJp"<<iterTask<<" = "<< Jac*Jp <<endl;
	sotDEBUG(45) << "U"<<iterTask<<" = "<< U<<endl;
	sotDEBUG(45) << "Svd"<<iterTask<<" = "<< S<<endl;
	sotDEBUG(45) << "V"<<iterTask<<" = "<< V<<endl;

	mem->jacobianInvSINOUT = Jp;
	mem->jacobianInvSINOUT.setTime( iterTime );
	mem->jacobianConstrainedSINOUT = Jac;
	mem->jacobianConstrainedSINOUT.setTime( iterTime );
	mem->singularBaseImageSINOUT = V;
	mem->singularBaseImageSINOUT.setTime( iterTime );
	mem->rankSINOUT = rankJ;
	mem->rankSINOUT.setTime( iterTime );

	sotDEBUG(25)<<"Inverse recomputed."<<endl;

      } else {
	sotDEBUG(15)<<"Inverse not recomputed."<<endl;
	rankJ = mem->rankSINOUT.accessCopy();
	Jac = mem->jacobianConstrainedSINOUT.accessCopy();
	Jp = mem->jacobianInvSINOUT.accessCopy();
	V = mem->singularBaseImageSINOUT.accessCopy();
      }

	/***/sotCOUNTER(6,7); // TRACE

      /* --- COMPUTE QDOT AND P --- */
      /*DEBUG*/// if( iterTask==0 ) control += Jp*err; else
      //Jp*err;
      sotDEBUG(15) << "G = " << Jp<<endl;
      sotDEBUG(15) << "e = " << err<<endl;
      sotDEBUG(15) << "Ge = " << Jp*err<<endl;
      sotDEBUG(15) << "d = " << Jac*control<<endl;
      control += Jp*(err - Jac*control);
      /***/sotCOUNTER(7,8); // QDOT


      if( mJ!=P.nbCols() ) { P.resize( mJ,mJ ); P.setIdentity(); }
      {
	ml::Matrix Kp( mJ,mJ );
	Jp.multiply( Jt,Kp );
	P -= Kp;
      }
      sotCOUNTER(8,9); // Projo

      sotDEBUG(15) << "q"<<iterTask<<" = "<<control<<std::endl;
      sotDEBUG(25) << "P"<<iterTask<<" = "<< P <<endl;
#ifdef WITH_CH
      sotDEBUG(1) << "Chrono"<<iterTask<<"_"<<iterTime<<" = [ " << endl
		  <<" "<<dt0b<<" "<<dt1 <<" "<< dt2 <<" "<< dt3 <<" "
		  << dt4 <<" "<<dt5 <<", "<< dt6 <<" "<< dt7
		  <<" "<< dt8 <<" "<< dt9 << " ] " << endl;
      sotDEBUG(1) << "ChronoKh"<<iterTask<<"_"<<iterTime<<" = [ " << endl
		  <<" "<< dt4a<<" "<< dt4b<<" "<< dt4c<<" "
		  << dt4d<<" "<< dt4e<<" " << " ] " << endl;
#endif // #ifdef WITH_CHRONO
      iterTask++;
    }

  sotCHRONO1 << endl;


  sotDEBUGOUT(15);
  return control;
}







